@@ -1,8 +1,10 @@
 import csv
 import datetime
 import random
+from operator import itemgetter
 
 import lightgbm as lgb
+import numpy as np
 import pandas as pd
 from catboost import CatBoostClassifier, CatBoostRegressor
 from sklearn.ensemble import (
@@ -13,8 +15,16 @@ from sklearn.ensemble import (
     RandomForestClassifier,
     RandomForestRegressor,
 )
-from sklearn.linear_model import ElasticNet, LogisticRegression
+from sklearn.feature_selection import SelectKBest, chi2
+from sklearn.linear_model import (
+    ElasticNet,
+    LogisticRegression,
+    SGDClassifier,
+    SGDRegressor,
+)
 from sklearn.model_selection import GridSearchCV
+from sklearn.multioutput import MultiOutputClassifier
+from sklearn.naive_bayes import MultinomialNB
 from sklearn.neural_network import MLPClassifier, MLPRegressor
 from sklearn.pipeline import Pipeline
 from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
@@ -330,3 +340,107 @@ class TrainModel:
                     csvwriter.writerow(row)
 
         return g.best_params_
+
+    def run_feature_selection(self, X_train, y_train, k, is_multiclass):
+        """
+        Run SelectKBest feature selection for given datasets.
+        Implements a custom method of feature selection for multiclass targets.
+
+        Parameters
+        --------
+        X_train: DataFrame
+        y_train: DataFrame
+        k: int
+        is_multiclass: bool
+
+        Returns
+        --------
+        cols_to_keep: list
+
+        """
+        if is_multiclass:
+            all_feature_scores = []
+            for label in y_train.columns:
+                # Select from all first, limit to k after means
+                selector = SelectKBest(chi2, k="all")
+                selector.fit(X_train, y_train[label])
+                all_feature_scores.append(list(selector.scores_))
+            # Get mean feature scoring across all target classes
+            mean_feature_scores = np.mean(all_feature_scores, axis=0)
+            # Remove nulls
+            no_nulls_scores = np.nan_to_num(mean_feature_scores)
+            # Sort descending values & keep top k
+            selected_feature_idxs = np.argsort(no_nulls_scores)[::-1][:k]
+            cols_to_keep = list(itemgetter(*selected_feature_idxs)(X_train.columns))
+        else:
+            selector = SelectKBest(k=k)
+            selector.fit(X_train, y_train)
+            cols_to_keep = list(
+                X_train.columns[np.where(selector.get_support() == True)].values
+            )
+
+        return cols_to_keep
+
+    def train_in_chunks(
+        self, X_train, y_train, model_type, is_multiclass, chunk_sizes=500000
+    ):
+        """
+        For large datasets, train model in managable chunk sizes
+
+        Parameters
+        --------
+        X_train: DataFrame
+        y_train: DataFrame
+        model_type: str
+            'classification' or 'regression'
+        is_multiclass: bool
+        chunk_sizes: int
+
+        Returns
+        --------
+        model: sklearn model
+        """
+        # Set default num_chunks to number needed to get manageable chunk sizes
+        # Used this number after testing time needed to train various sizes
+        num_chunks = round(y_train.shape[0] / chunk_sizes)
+        # Make sure num_chunks isn't below 1
+        if num_chunks < 1:
+            num_chunks = 1
+
+        step = round(y_train.shape[0] / num_chunks)
+        idx_start = 0
+        idx_end = step
+
+        if is_multiclass:
+            # Need this custom model to do partial_fit with multiclass data
+            model = MultiOutputClassifier(MultinomialNB())
+        elif model_type == "classification":
+            model = SGDClassifier()
+        else:
+            model = SGDRegressor()
+
+        while idx_start < y_train.shape[0]:
+            if is_multiclass:
+                model.partial_fit(
+                    X_train.iloc[idx_start:idx_end],
+                    y_train.iloc[idx_start:idx_end],
+                    classes=[
+                        np.array(np.unique(y_train[col])) for col in y_train.columns
+                    ],
+                )
+            elif model_type == "classification":
+                model = model.partial_fit(
+                    X_train.iloc[idx_start:idx_end],
+                    y_train.iloc[idx_start:idx_end],
+                    classes=np.array(np.unique(y_train)),
+                )
+            else:
+                model = model.partial_fit(
+                    X_train.iloc[idx_start:idx_end],
+                    y_train.iloc[idx_start:idx_end],
+                )
+
+            idx_start += step
+            idx_end += step
+
+        return model
